/*
 * Copyright (c) 2001-2015, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH2-EA software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licensing information, please refer to the
 * copyright file COPYRIGHT in the top level MVAPICH2 directory.
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>

#include "koemt.h"

#ifdef _MG_DEBUG
#define dprintk(args...) printk(args);
#else
#define dprintk(args...)
#endif

MODULE_AUTHOR("Khaled Hamidouche <hamidouc@cse.ohio-state.edu>");
MODULE_DESCRIPTION("koemt: linux Kernel module for OSU Energy Monitoring Tool");
MODULE_VERSION("0.8");
MODULE_LICENSE("Dual BSD/GPL");


static int koemt_open(struct inode* i, struct file* f)
{
    return 0;
}

static int koemt_release(struct inode* i, struct file* f)
{
    return 0;
}

static ssize_t koemt_read(struct file *f, char *b, size_t c, loff_t *o)
{
    return 0;
}

static ssize_t koemt_write(struct file *f, const char *b, size_t c, loff_t *o)
{
    return 0;
}

static long koemt_ioctl(struct file *f, unsigned int ioctl_num, unsigned long ioctl_param);

dev_t koemt_dev;
struct cdev *koemt_cdev;

struct file_operations koemt_fops = {
    .owner =          THIS_MODULE,
    .read =           koemt_read,
    .write =          koemt_write,
    .open =           koemt_open,
    .release =        koemt_release,
    .unlocked_ioctl = koemt_ioctl,
    .compat_ioctl =   NULL,
};

static long long read_msr(unsigned int ecx) {

    if (!(ecx==MSR_RAPL_POWER_UNIT || ecx== MSR_PKG_ENERGY_STATUS || 
            ecx==MSR_PP0_ENERGY_STATUS || ecx== MSR_DRAM_ENERGY_STATUS )){
            return -1;
    }

    unsigned int edx = 0, eax = 0;
    unsigned long long result = 0;
    __asm__ __volatile__("rdmsr" : "=a"(eax), "=d"(edx) : "c"(ecx));
    result = eax | (unsigned long long)edx << 0x20;
    dprintk(KERN_ALERT "Module msrdrv: Read 0x%016llx (0x%08x:0x%08x) from MSR 0x%08x\n", result, edx, eax, ecx)
    return result;
}



static long koemt_ioctl(struct file *f, unsigned int ioctl_num, unsigned long ioctl_param)
{
    struct MsrInOut *msrops;
    int i;
    if (ioctl_num != IOCTL_MSR_CMDS) {
            return 0;
    }
    msrops = (struct MsrInOut*)ioctl_param;
    for (i = 0 ; i <= MSR_VEC_LIMIT ; i++, msrops++) {
        switch (msrops->op) {
        case MSR_NOP:
            dprintk(KERN_ALERT "Module " DEV_NAME ": seen MSR_NOP command\n")
            break;
        case MSR_STOP:
            dprintk(KERN_ALERT "Module " DEV_NAME ": seen MSR_STOP command\n")
            goto label_end;
        case MSR_READ:
            dprintk(KERN_ALERT "Module " DEV_NAME ": seen MSR_READ command\n")
            msrops->value = read_msr(msrops->ecx);
            break;
        default:
            dprintk(KERN_ALERT "Module " DEV_NAME ": Unknown option 0x%x\n", msrops->op)
            return 1;
        }
    }
    label_end:

    return 0;
}


static int koemt_init(void)
{
    koemt_dev = MKDEV(DEV_MAJOR, DEV_MINOR);
    register_chrdev_region(koemt_dev, 1, DEV_NAME);
    koemt_cdev = cdev_alloc();
    koemt_cdev->owner = THIS_MODULE;
    koemt_cdev->ops = &koemt_fops;
    cdev_init(koemt_cdev, &koemt_fops);
    cdev_add(koemt_cdev, koemt_dev, 1);
    printk(KERN_ALERT "Module " DEV_NAME " loaded\n");

    return 0;
}

static void koemt_exit(void)
{
    cdev_del(koemt_cdev);
    unregister_chrdev_region(koemt_dev, 1);
    printk(KERN_ALERT "Module " DEV_NAME " unloaded\n");
}

module_init(koemt_init);
module_exit(koemt_exit);

